/* MLIST.C: Read the master list. */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "romdb.h"

/*************************************************************************
 * Constants
 *************************************************************************/

static const unsigned char CC2_HR = 0; /* horizontal rule character */

/*************************************************************************
 * Structures
 *************************************************************************/

struct romdb
{
	size_t size;	/* number of entries */
	struct dbnode* first;
	struct dbnode* last;
};

struct dbnode
{
	/* "data" is first to enable quick casting from (dbnode*) to
	 * (romentry*) */
	struct romentry data;
	struct dbnode* next;
};

/*************************************************************************
 * Prototypes
 *************************************************************************/

static void makeupper(char *string);
static void munge_filename(char *string);
static int isValidFATName(const char *s);
static size_t fgettok(char *buf, size_t maxlen, const char *delim, FILE *f);
static size_t fskipspn(const char *accept, FILE *f);
static size_t fskipcspn(const char *reject, FILE *f);
static int romentry_report_errors(const struct romentry *ent);

/*************************************************************************
 * Global Variables
 *************************************************************************/

struct romentry** RomList = NULL;

/**
 * Convert string to upper case.
 */
static void makeupper(char *string)
{
	size_t i;
	for(i = 0; i < strlen(string); i++)
	{
		string[i] = toupper(string[i]);
	}
}

/**
 * Process a file name to make it work well on CC2.
 */
static void munge_filename(char *string)
{
	/* Allow a minus sign to mark file name fields without a value. */
	if (0 == strcmp(string, "-"))
	{
		string[0] = 0;
	}
	else
	{
		makeupper(string);
	}
}

/**
 * Check that a file name is good for DOS.
 */
static int isValidFATName(const char *s)
{
	static const char FATCHARS[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";
	size_t first_bad_char = strspn(s, FATCHARS);
	return s[first_bad_char] == '\0';
}

/**
 * Consume characters from a file and store them in "buf".  Consume and discard
 * characters beyond "maxlen" length.  Stop when a character in the set "delim"
 * is encountered.
 *
 * Returns number of chars stored in buffer (0 on EOF).
 */
static size_t fgettok(char *buf, size_t maxlen, const char *delim, FILE *f)
{
	size_t n = 0;
	int c;

	c = fgetc(f);
	if (c == EOF)
		goto done;

	while (n < maxlen)
	{
		if (strchr(delim, c))
			break;

		buf[n++] = c;

		c = fgetc(f);
		if (c == EOF)
			goto done;
	}

	if (n == maxlen) /* discard extra text characters */
	{
		while (!strchr(delim, c))
		{
			c = fgetc(f);
			if (c == EOF)
				goto done;
		}
	}

	ungetc(c, f);

 done:
	buf[n] = 0;
	return n;
}

/**
 * Consume characters from a file until one not in the set "accept" is
 * encountered.
 *
 * Returns number of chars skipped (0 on EOF).
 */
static size_t fskipspn(const char *accept, FILE *f)
{
	size_t n = 0;
	int c;

	do
	{
		c = fgetc(f);
		if (c == EOF)
			goto done;
		++n;

	} while (strchr(accept, c));

	ungetc(c, f);
	--n;

 done:
	return n;
}

/**
 * Consume characters from a file until one in the set "reject" is encountered.
 *
 * Returns number of chars skipped (0 on EOF).
 */
static size_t fskipcspn(const char *reject, FILE *f)
{
	size_t n = 0;
	int c;

	do
	{
		c = fgetc(f);
		if (c == EOF)
			goto done;
		++n;

	} while (!strchr(reject, c));

	ungetc(c, f);
	--n;

 done:
	return n;
}

/**
 * Scan a database entry for errors and report them.
 *
 * Returns the number of errors found.
 */
static int romentry_report_errors(const struct romentry *ent)
{
	int error_count = 0;
	int rom_valid = 0;
	int bs_valid = 0;
	int st_valid = 0;

	if (ent->RomName[0])
	{
		if (!isValidFATName(ent->RomName))
		{
			fprintf(stderr,
				"Error: invalid ROM file name: %s\n",
				ent->RomName);
			++error_count;
		}
		else
		{
			rom_valid = 1;
		}
	}

	if (ent->Bankswitch[0])
	{
		if (!isValidFATName(ent->Bankswitch))
		{
			fprintf(stderr,
				"Error: invalid bankswitch file name: %s\n",
				ent->Bankswitch);

			++error_count;
		}
		else
		{
			bs_valid = 1;
		}
	}

	if (ent->Startup[0])
	{
		if (!isValidFATName(ent->Startup))
		{
			fprintf(stderr,
				"Error: invalid startup file name: %s\n",
				ent->Startup);
			++error_count;
		}
		else
		{
			st_valid = 1;
		}
	}

	if (bs_valid)
	{
		if (!rom_valid)
		{
			fprintf(stderr,
			"Error: bankswitch file without ROM file: %s\n",
				ent->Bankswitch);
				++error_count;
		}

		if (!st_valid)
		{
			fprintf(stderr,
			"Error: bankswitch file without startup file: %s\n",
				ent->Bankswitch);
			++error_count;
		}
	}

	if (st_valid)
	{
		if (!rom_valid)
		{
			fprintf(stderr,
				"Error: startup file without ROM file: %s\n",
				ent->Startup);
				++error_count;
		}

		if (!bs_valid)
		{
			fprintf(stderr,
			"Error: startup file without bankswitch file: %s\n",
				ent->Startup);
			++error_count;
		}

	}

#if 0 /* Is this good to be here or not? */

	/* Supercharger loads (identified by a ROM with no bankswitch
	 * or startup files) must have len=8 and end in 2 digits. */
	if (ent->RomName[0] && !ent->Bankswitch[0] && !ent->Startup[0])
	{
		if (strlen(ent->RomName < MAX_FILE_NAME_LEN)
		    || !isdigit(ent->RomName[MAX_FILE_NAME_LEN-2])
		    || !isdigit(ent->RomName[MAX_FILE_NAME_LEN-1]))
		{
			/* TO DO: report error */
			++error_count;
		}
	}
#endif

	return error_count;
}

/**
 * Reads the given file and returns a ROM database that the caller
 * must free.
 */
struct romdb* romdb_load(FILE *file)
{
	static const char delim[] = "\t\r\n";
	static const char tab[] = "\t";
	static const char eol[] = "\r\n";

	struct romdb* db = NULL;
	struct romentry* ent = NULL;

	db = malloc(sizeof(*db));
	if (db)
	{
		memset(db, 0, sizeof(*db));

		/* N.B. Although ent is a romentry*, allocate space
		 * for a dbnode, so that it can be added to the db
		 * later. */
		ent = malloc(sizeof(struct dbnode));
		if (ent)
		{
			memset(ent, 0, sizeof(struct dbnode));
			goto got_mem;
		}

		free(db);
		db = NULL;
	}

	goto done;

 got_mem:

	while (fgettok(ent->Name, MAX_GAME_NAME_LEN, delim, file) > 0)
	{
		char crcstr[9];
		char dummy;
		int error_count = 0;

		ent->RomName[0] = 0;
		ent->Bankswitch[0] = 0;
		ent->Startup[0] = 0;
		ent->CRC = 0;

		fskipspn(tab, file);	/* consume tabs */
		if (fgettok(ent->RomName,
			    MAX_FILE_NAME_LEN, delim, file) == 0)
		{
			goto line_done;
		}
		munge_filename(ent->RomName);

		fskipspn(tab, file);	/* consume tabs */
		if (fgettok(ent->Bankswitch,
			    MAX_FILE_NAME_LEN, delim, file) == 0)
		{
			goto line_done;
		}
		munge_filename(ent->Bankswitch);

		fskipspn(tab, file);	/* consume tabs */
		if (fgettok(ent->Startup,
			    MAX_FILE_NAME_LEN, delim, file) == 0)
		{
			goto line_done;
		}
		munge_filename(ent->Startup);

		fskipspn(tab, file);	/* consume tabs */
		if (fgettok(crcstr, 8, delim, file) == 0)
		{
			goto line_done;
		}

	line_done:

		fskipcspn(eol, file);	/* discard chars to end of line */
		fskipspn(eol, file);	/* discard end of line */

		if (sscanf(crcstr, "%lx%c", &ent->CRC, &dummy) != 1)
		{
			fprintf(stderr, "Error: invalid CRC: %s\n", crcstr);
			++error_count;
		}

		error_count += romentry_report_errors(ent);

		if (error_count > 0)
		{
			continue; /* discard this line and keep going */
		}

		if (!ent->RomName[0])
		{
			/* If this were MENU.TXT, we would write out a
			 * horizonal rule. */
			continue;
		}
		else
		{
			/* Finally, add the entry to the db.  Note
			 * that enough space was allocated for ent to
			 * be cast to a dbnode. */
			struct dbnode* node = (struct dbnode*) ent;
			node->next = NULL;
			if (db->last)
			{
				db->last->next = node;
			}
			db->last = node;
			if (!db->first)
			{
				db->first = node;
			}
			db->size += 1;

			/* N.B. Although ent is a romentry*, allocate space
			 * for a dbnode, so that it can be added to the db
			 * later. */
			ent = malloc(sizeof(struct dbnode));
			if (ent)
			{
				memset(ent, 0, sizeof(struct dbnode));
			}
			else
			{
				/* TO DO: print error? */
				free(db);
				goto done;
			}
		}
	}

 done:
	if (ent)
	{
		free(ent);
		ent = NULL;
	}


	return db;
}

int romdb_size(const struct romdb* db)
{
	return db->size;
}

struct romentry* romdb_find_crc(struct romdb* db, unsigned long crc, unsigned long *count)
{
	struct dbnode* node;

	*count = 0;

	for (node = db->first; node; node = node->next)
	{
		if (crc == node->data.CRC)
		{
			return &node->data;
		}
		(*count)++;
	}

	return NULL;
}

int romdb_walk(struct romdb* db, romdb_walk_fn* fn)
{
	struct dbnode* node = db->first;
	int result = 0;

	while (node && (result >= 0))
	{
		/* romdb_free uses this to delete nodes as it walks,
		 * so follow the link before calling fn(). */
		struct dbnode* current = node;
		node = node->next;
		result = fn(&current->data);
	}

	return result;
}

void romdb_free(struct romdb* db)
{
	romdb_walk(db, (romdb_walk_fn*)free);
	free(db);
}
